@@ -1,8 +1,11 @@
 import os
+import glob
 import random
 import json
 from contextlib import contextmanager
 from http.server import SimpleHTTPRequestHandler
+import shutil
+import pathlib
 import socketserver
 import socket
 import logging
@@ -27,20 +30,17 @@ def working_directory(path):
     """
     current_directory=os.getcwd()
     os.chdir(path)
+    log.debug(f"Now working in directory {path}")
     try:
         yield
     finally:
         os.chdir(current_directory)
+        log.debug(f"Successfully changed directory back to {current_directory}")
 
-
-def ensure_directory(path):
-    """
-    If the directory doesn't exist yet, create it.
-    """
-    try:
-        os.makedirs(path)
-    except FileExistsError:
-        pass
+def linux_path(path):
+    # hack to make core ptx and xsl:import happy
+    p = pathlib.Path(path)
+    return p.as_posix()
 
 
 def directory_exists(path):
@@ -70,6 +70,9 @@ def project_xml(dirpath=os.getcwd()):
         project_manifest = os.path.join(project_path(dirpath), 'project.ptx')
     return ET.parse(project_manifest)
 
+def project_xml_string(dirpath=os.getcwd()):
+    return ET.tostring(project_xml(dirpath), encoding='unicode')
+
 def target_xml(alias=None,dirpath=os.getcwd()):
     if alias is None:
         return project_xml().find("targets/target")
@@ -103,18 +106,14 @@ def xml_syntax_is_valid(xmlfile):
     except IOError:
         log.error(f'The file {xmlfile} does not exist')
         return False
-
     # check for XML syntax errors
     except ET.XMLSyntaxError as err:
-        log.error('XML Syntax Error, see error_syntax.log. Quitting...')
-        with open('error_syntax.log', 'w') as error_log_file:
-            error_log_file.write(str(err.error_log))
+        log.error('XML Syntax Error caused build to fail:')
+        log.error(str(err.error_log))
         return False
     except ET.XIncludeError as err:
-        log.error(
-            'XML Syntax Error with instance of xinclude; see error_syntax.log. Quitting...')
-        with open('error_syntax.log', 'w') as error_log_file:
-            error_log_file.write(str(err.error_log))
+        log.error('XInclude Error caused build to fail:')
+        log.error(str(err.error_log))
         return False
     return True
 
@@ -150,11 +149,15 @@ class HTMLRebuildHandler(watchdog.events.FileSystemEventHandler):
         self.last_trigger_at = time.time()-5
         self.callback = callback
     def on_any_event(self,event):
-        # only trigger at most every 5 seconds
-        if time.time() > self.last_trigger_at + 5:
-            self.last_trigger_at = time.time()
-            log.info(f"\nChanges to source detected.\n")
-            self.callback()
+        self.last_trigger_at = time.time()
+        # only run callback once triggers halt for a second
+        def timeout_callback(handler):
+            time.sleep(1.5)
+            if time.time() > handler.last_trigger_at + 1:
+                handler.last_trigger_at = time.time()
+                log.info("\nChanges to source detected.\n")
+                handler.callback()
+        threading.Thread(target=timeout_callback,args=(self,)).start()
 
 # boilerplate to prevent overzealous caching by preview server, and
 # avoid port issues
@@ -193,7 +196,8 @@ def serve_forever(directory,access="private",port=8000):
             with TCPServer((binding, port), RequestHandler) as httpd:
                 looking_for_port = False
                 url = url_for_access(access,port)
-                log.info(f"Server started at {url}\n")
+                log.info(f"Success! Open the below url in a web browser to preview the most recent build of your project.")
+                log.info("    "+url)
                 log.info("Use [Ctrl]+[C] to halt the server.\n")
                 httpd.serve_forever()
         except OSError:
@@ -231,4 +235,27 @@ def expand_pretext_href(lxml_element):
     Expands @pretext-href attributes to point to the distributed xsl directory.
     '''
     for ele in lxml_element.xpath('//*[@pretext-href]'):
-        ele.set('href',str(static.core_xsl(ele.get('pretext-href'),as_path=True)))
+        ele.set('href',str(linux_path(static.core_xsl(ele.get('pretext-href'),as_path=True))))
+
+def copy_expanded_xsl(xsl_path: str, output_dir: str):
+    """
+    Copy relevant files that share a directory with `xsl_path`
+    while pre-processing the `.xsl` files.
+    """
+    xsl_dir = os.path.abspath(os.path.dirname(xsl_path))
+    output_dir = os.path.abspath(output_dir)
+    log.debug(f"Copying all files in {xsl_dir} to {output_dir}")
+    shutil.copytree(xsl_dir, output_dir, dirs_exist_ok=True)
+    # expand each xsl file
+    with working_directory(output_dir):
+        for filename in glob.iglob('**',recursive=True):
+            # glob lists both files and directories, but we only want to copy files.
+            if os.path.isfile(filename) and filename.endswith('.xsl'):
+                log.debug(f"Expanding and copying {filename}")
+                try:
+                    lxml_element = ET.parse(filename)
+                    expand_pretext_href(lxml_element)
+                    lxml_element.write(filename)
+                # maybe an xsl file is malformed, but let's continue in case it's unused
+                except Exception as e:
+                    log.warning(f"Hit error `{e}` when expanding {filename}, continuing anyway...")
